@@ -16,13 +16,14 @@ limitations under the License.
 
 import json
 
-from .serviceconnector import ServiceConnector
+from .serviceconnector import _Client
 from .utils import get_logger
+from .utils import raise_for_status_with_detail
 
 log = get_logger(__name__)
 
 
-class AuthenticationClient:
+class AuthenticationClient(_Client):
     """
     Client authentication.
     """
@@ -32,9 +33,6 @@ class AuthenticationClient:
             'upgrade':      'accounts/token/upgrade'
            }
 
-    def __init__(self, url, version):
-        self._serviceconnector = ServiceConnector(url, version, None)
-
     def fetch_auth_token(self, tenant_id, username, password):
         """
         Retrieves the JWT token for a given user in a given tenant.
@@ -50,7 +48,7 @@ class AuthenticationClient:
         body_s = json.dumps(body)
         headers = {'Content-Type': 'application/json'}
         res = self._serviceconnector.request('POST', uri, body_s, headers)
-        res.raise_for_status()
+        raise_for_status_with_detail(res)
         return res.json()['jwt']
 
     def register(self, tenant_info, invitation_code):
@@ -65,5 +63,5 @@ class AuthenticationClient:
         headers = {'Content-Type': 'application/json'}
         params = {'invitationCode': invitation_code}
         res = self._serviceconnector.request('POST', uri, body_s, headers, params=params)
-        res.raise_for_status()
+        raise_for_status_with_detail(res)
         return res.json()
